@@ -6,16 +6,13 @@ use Carp ();
 ############################################################################
 package Net::OpenID::Server;
 
-use vars qw($VERSION $HAS_CRYPT_DSA $HAS_CRYPT_OPENSSL);
-$VERSION = "0.05";
+use vars qw($VERSION);
+$VERSION = "0.06";
 
 use fields (
             'last_errcode',   # last error code we got
             'last_errtext',   # last error code we got
 
-            'private_key',     # filename, scalar, or subref returning PEM contents
-            'public_key',      # filename, scalar, or subref returning PEM contents
-
             'get_user',        # subref returning a defined value representing the logged in user, or undef if no user.
                                # this return value ($u) is passed to the other subrefs
 
@@ -36,7 +33,8 @@ use fields (
                                # setup_url to your preferred names.  the standard keys are:
                                # "trust_root", "return_to", "post_grant", and "is_identity"
 
-            'args',            # that magic anything reference (returns a subref)
+            'get_args',        # thing to get get args
+            'post_args',       # thing to get post args
 
             'server_secret',    # subref returning secret given $time
             'secret_gen_interval',
@@ -48,14 +46,7 @@ use MIME::Base64 ();
 use Digest::SHA1 qw(sha1 sha1_hex);
 use Crypt::DH 0.05;
 use Math::BigInt;
-
-BEGIN {
-    unless ($HAS_CRYPT_OPENSSL = eval "use Crypt::OpenSSL::DSA 0.12; 1;") {
-        unless ($HAS_CRYPT_DSA = eval "use Crypt::DSA (); use Convert::PEM 0.07; 1;") {
-            warn "Old protocol mode requires Crypt::OpenSSL::DSA version 0.12+ -or- (Crypt::DSA -and- Convert::PEM 0.07+)\n";
-        }
-    }
-}
+use Time::Local qw(timegm);
 
 sub new {
     my Net::OpenID::Server $self = shift;
@@ -65,14 +56,14 @@ sub new {
     $self->{last_errcode} = undef;
     $self->{last_errtext} = undef;
 
-    $self->args(delete $opts{args});
+    $self->get_args(delete $opts{get_args} || delete $opts{args});
+    $self->post_args(delete $opts{post_args});
 
     $opts{'secret_gen_interval'} ||= 86400;
     $opts{'secret_expire_age'}   ||= 86400 * 14;
 
     $self->$_(delete $opts{$_})
         foreach (qw(
-                    private_key public_key
                     get_user is_identity is_trusted
                     setup_url setup_map server_secret
                     secret_gen_interval secret_expire_age
@@ -86,9 +77,6 @@ sub get_user     { &_getsetcode; }
 sub is_identity  { &_getsetcode; }
 sub is_trusted   { &_getsetcode; }
 
-sub public_key  { &_getsetkey; }
-sub private_key { &_getsetkey; }
-
 sub setup_url   { &_getset; }
 sub setup_map   { &_getset; }
 
@@ -111,15 +99,20 @@ sub handle_page {
     Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
     Carp::croak("handle_page must be called in list context") unless wantarray;
 
-    # OLDPROTO:
-    return $self->_mode_pubkey
-        if $self->args("openid.mode") eq "getpubkey";
-
     return $self->_mode_associate
-        if $self->args("openid.mode") eq "associate";
+        if $self->pargs("openid.mode") eq "associate";
 
-    my $mode = $self->args("openid.mode") || "checkid_immediate";
-    return $self->_fail("unknown_mode") unless $mode =~ /^checkid_(?:immediate|setup)/;
+    return $self->_mode_check_authentication
+        if $self->pargs("openid.mode") eq "check_authentication";
+
+    my $mode = $self->args("openid.mode");
+    unless ($mode) {
+        return ("text/html",
+                "<html><head><title>OpenID Endpoint</title></head><body>This is an OpenID server endpoint, not a human-readable resource.  For more information, see <a href='http://openid.net/'>http://openid.net/</a>.</body></html>");
+    }
+
+    return $self->_error_page("Unknown mode")
+        unless $mode =~ /^checkid_(?:immediate|setup)/;
 
     return $self->_mode_checkid($mode, $redirect_for_setup);
 }
@@ -134,15 +127,17 @@ sub handle_page {
 
 #   ...
 
-sub args {
+# GET args
+*args = \&get_args;
+sub get_args {
     my Net::OpenID::Server $self = shift;
 
     if (my $what = shift) {
         Carp::croak("Too many parameters") if @_;
         my $getter;
         if (! ref $what){
-            Carp::croak("No args defined") unless $self->{args};
-            return $self->{args}->($what) || "";
+            Carp::croak("No get_args defined") unless $self->{get_args};
+            return $self->{get_args}->($what) || "";
         } elsif (ref $what eq "HASH") {
             $getter = sub { $what->{$_[0]}; };
         } elsif (ref $what eq "CGI") {
@@ -158,48 +153,39 @@ sub args {
             Carp::croak("Unknown parameter type ($what)");
         }
         if ($getter) {
-            $self->{args} = $getter;
+            $self->{get_args} = $getter;
         }
     }
-    $self->{args};
+    $self->{get_args};
 }
 
-# OLDPROTO: will be removed
-sub signed_return {
+# POST args
+*pargs = \&post_args;
+sub post_args {
     my Net::OpenID::Server $self = shift;
-    my %opts = @_;
-    my $identity   = delete $opts{'identity'};
-    my $return_to  = delete $opts{'return_to'};
-    my $trust_root = delete $opts{'trust_root'};  # optional
-    Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
 
-    if ($trust_root) {
-        return undef unless _url_is_under($trust_root, $return_to);
+    if (my $what = shift) {
+        Carp::croak("Too many parameters") if @_;
+        my $getter;
+        if (! ref $what){
+            Carp::croak("No pargs defined") unless $self->{post_args};
+            return $self->{post_args}->($what) || "";
+        } elsif (ref $what eq "HASH") {
+            $getter = sub { $what->{$_[0]}; };
+        } elsif (ref $what eq "CGI") {
+            $getter = sub { scalar $what->param($_[0]); };
+        } elsif (ref $what eq "Apache::Request") {
+            $getter = sub { scalar $what->param($_[0]); };
+        } elsif (ref $what eq "CODE") {
+            $getter = $what;
+        } else {
+            Carp::croak("Unknown parameter type ($what)");
+        }
+        if ($getter) {
+            $self->{post_args} = $getter;
+        }
     }
-
-    my $now = _time_to_w3c();
-    my $plain = join("::",
-                     $now,
-                     "assert_identity",
-                     $identity,
-                     $return_to);
-
-    my $sig = $self->_dsa_sig($plain)
-        or return;
-
-    die "Failed to make signature (has length = " . length($sig) . ")"
-        unless length($sig) >= 45 && length($sig) <= 48;
-
-    my $sig64 = MIME::Base64::encode_base64($sig);
-    chomp $sig64;  # remove \n
-    my $ret_url = $return_to;
-    _push_url_arg(\$ret_url,
-                  "openid.mode",            "id_res",
-                  "openid.assert_identity", $identity,
-                  "openid.sig",             $sig64,
-                  "openid.timestamp",       $now,
-                  "openid.return_to",       $return_to);
-    return $ret_url;
+    $self->{post_args};
 }
 
 sub signed_return_url {
@@ -208,7 +194,7 @@ sub signed_return_url {
     my $identity     = delete $opts{'identity'};
     my $return_to    = delete $opts{'return_to'};
     my $assoc_handle = delete $opts{'assoc_handle'};
-    my $assoc_type   = delete $opts{'assoc_type'} || "HMAC-SHA1";
+    my $assoc_type   = "HMAC-SHA1";  # TODO: function to get type given a handle
     my $trust_root   = delete $opts{'trust_root'};  # optional
     Carp::croak("Unknown options: " . join(", ", keys %opts)) if %opts;
 
@@ -216,7 +202,6 @@ sub signed_return_url {
         return undef unless _url_is_under($trust_root, $return_to);
     }
 
-    # new protocol
     my $ret_url = $return_to;
 
     my $c_sec;
@@ -234,10 +219,9 @@ sub signed_return_url {
     my %arg = (
                mode         => "id_res",
                issued       => _time_to_w3c($now),
-               valid_to     => _time_to_w3c($now + 30),
+               valid_to     => _time_to_w3c($now + 60),  # TODO: make 60 configurable?
                identity     => $identity,
                return_to    => $return_to,
-               assoc_type   => $assoc_type,
                assoc_handle => $assoc_handle,
                signed       => join(",", @sign),
                );
@@ -273,16 +257,6 @@ sub _mode_checkid {
 
     my $identity = $self->args("openid.identity");
 
-    # OLDPROTO:
-    my $old_proto = 0;
-    unless ($identity) {
-        if ($identity = $self->args("openid.is_identity")) {
-            $old_proto = 1;
-        } else {
-            return $self->_fail("no_identity");
-        }
-    }
-
     # chop off the query string, in case our trust_root came from the return_to URL
     $trust_root =~ s/\?.*//;
 
@@ -292,20 +266,11 @@ sub _mode_checkid {
 
     # assertion path:
     if ($is_identity && $is_trusted) {
-        my $ret_url;
-        if ($old_proto) {
-            $ret_url = $self->signed_return(
+        my $ret_url = $self->signed_return_url(
                                                identity => $identity,
                                                return_to => $return_to,
+                                               assoc_handle => $self->args("openid.assoc_handle"),
                                                );
-        } else {
-            $ret_url = $self->signed_return_url(
-                                                identity => $identity,
-                                                return_to => $return_to,
-                                                assoc_handle => $self->args("openid.assoc_handle"),
-                                                assoc_type   => $self->args("openid.assoc_type"),
-                                                );
-        }
         return ("redirect", $ret_url);
     }
 
@@ -314,9 +279,10 @@ sub _mode_checkid {
     # with a setup URL (the default), or explictly said that we're in control of
     # the user-agent's full window, and we can do whatever we want with them now.
     my %setup_args = (
-                      $self->_setup_map("trust_root"),  $trust_root,
-                      $self->_setup_map("return_to"),   $return_to,
-                      $self->_setup_map("identity"),    $identity,
+                      $self->_setup_map("trust_root"),   $trust_root,
+                      $self->_setup_map("return_to"),    $return_to,
+                      $self->_setup_map("identity"),     $identity,
+                      $self->_setup_map("assoc_handle"), $self->args("openid.assoc_handle"),
                       );
 
     my $setup_url = $self->{setup_url} or Carp::croak("No setup_url defined.");
@@ -333,6 +299,9 @@ sub _mode_checkid {
 
         if ($redirect_for_setup) {
             # ... unless the caller just wants us to redirect to their setup page
+            # and in that case we add post_grant=return since that's implied
+            # in the checkid_setup mode
+            _push_url_arg(\$setup_url, "openid.post_grant", "return");
             return ("redirect", $setup_url);
         } else {
             return ("setup", \%setup_args);
@@ -359,7 +328,7 @@ sub _proxy {
     return $getter->(@_);
 }
 
-sub get_server_secret {
+sub _get_server_secret {
     my Net::OpenID::Server $self = shift;
     my $time = shift;
 
@@ -386,7 +355,7 @@ sub _generate_association {
     my $now = time();
     my $sec_time = $now - ($now % $self->secret_gen_interval);
 
-    my $s_sec = $self->get_server_secret($sec_time)
+    my $s_sec = $self->_get_server_secret($sec_time)
         or Carp::croak("server_secret didn't return a secret given what should've been a valid time ($sec_time)\n");
 
     my $nonce = _rand_chars(20);
@@ -408,7 +377,7 @@ sub _secret_of_handle {
     my ($time, $nonce, $nonce_sig80) = split(/:/, $handle);
 
     my $sec_time = $time - ($time % $self->secret_gen_interval);
-    my $s_sec = $self->get_server_secret($sec_time)  or return;
+    my $s_sec = $self->_get_server_secret($sec_time)  or return;
     length($nonce)       == 20 or return;
     length($nonce_sig80) == 10 or return;
     return unless $no_verify || $nonce_sig80 eq substr(hmac_sha1_hex("$time:$nonce", $s_sec), 0, 10);
@@ -416,14 +385,6 @@ sub _secret_of_handle {
     return hmac_sha1($handle, $s_sec);
 }
 
-sub _mode_pubkey {
-    my Net::OpenID::Server $self = shift;
-    my $pubkey = $self->_get_key_contents("public")
-        or return;
-
-    return ("text/plain", $pubkey);
-}
-
 sub _mode_associate {
     my Net::OpenID::Server $self = shift;
 
@@ -446,14 +407,14 @@ sub _mode_associate {
     $prop{'expiry'}       = _time_to_w3c($expires);
     $prop{'issued'}       = _time_to_w3c($now);
 
-    if ($self->args("openid.enc") eq "DH-SHA1") {
+    if ($self->pargs("openid.session_type") eq "DH-SHA1") {
 
         my $dh   = Crypt::DH->new;
-        my $p    = _arg2bi($self->args("openid.dh_modulus"));
-        my $g    = _arg2bi($self->args("openid.dh_gen"));
-        my $cpub = _arg2bi($self->args("openid.dh_gx"));
+        my $p    = _arg2bi($self->pargs("openid.dh_modulus")) || _default_p();
+        my $g    = _arg2bi($self->pargs("openid.dh_gen"))     || _default_g();
+        my $cpub = _arg2bi($self->pargs("openid.dh_consumer_public"));
 
-        return $self->_error_page("invalid dh params")
+        return $self->_error_page("invalid dh params p=$p, g=$g, cpub=$cpub")
             unless $p > 10 && $g > 1 && $cpub;
 
         $dh->p($p);
@@ -462,26 +423,59 @@ sub _mode_associate {
 
         my $dh_sec = $dh->compute_secret($cpub);
 
-        $prop{'dh_gy'}      = _bi2arg($dh->pub_key);
-        $prop{'enc'}        = "DH-SHA1";
-        $prop{'enc_secret'} = _b64(sha1($secret) ^ sha1(_bi2bytes($dh_sec)));
+        $prop{'dh_server_public'} = _bi2arg($dh->pub_key);
+        $prop{'session_type'}     = "DH-SHA1";
+        $prop{'enc_mac_key'}      = _b64($secret ^ sha1(_bi2bytes($dh_sec)));
 
     } else {
-        $prop{'secret'} = _b64($secret);
+        $prop{'mac_key'} = _b64($secret);
     }
 
     return $self->_serialized_props(\%prop);
 }
 
+sub _mode_check_authentication {
+    my Net::OpenID::Server $self = shift;
+
+    my $signed = $self->pargs("openid.signed") || "";
+    my $token = "";
+    foreach my $param (split(/,/, $signed)) {
+        next unless $param =~ /^\w+$/;
+        my $val = $param eq "mode" ? "id_res" : $self->pargs("openid.$param");
+        next unless defined $val;
+        next if $val =~ /\n/;
+        $token .= "$param:$val\n";
+    }
+
+    my $sig = $self->pargs("openid.sig");
+    my $ahandle = $self->pargs("openid.assoc_handle")
+        or return $self->error_page("no_assoc_handle");
+
+    my $c_sec = $self->_secret_of_handle("HMAC-SHA1", $ahandle)
+        or return $self->_error_page("bad_handle");
+
+    my $good_sig = _b64(hmac_sha1($token, $c_sec));
+
+    my $lifetime = 0;
+    if ($sig eq $good_sig) {
+        my $valid_to = _w3c_to_time($self->pargs("openid.valid_to"));
+        my $now = time;
+        $lifetime = $valid_to - $now - 1;   # assume we took a second
+        $lifetime = 0 if $lifetime < 0;
+    }
+
+    return $self->_serialized_props({ 'lifetime' => $lifetime });
+}
+
 sub _b64 {
     my $val = MIME::Base64::encode_base64($_[0]);
-    chomp $val;
+    $val =~ s/\s+//g;
     return $val;
 }
 
 sub _error_page {
     my Net::OpenID::Server $self = shift;
-    return $self->_serialized_props({ 'error_text' => $_[0] });
+    return $self->_serialized_props({ 'error' => $_[0] });
 }
 
 sub _serialized_props {
@@ -549,18 +543,6 @@ sub _getsetcode {
     return $self->{$param};
 }
 
-sub _getsetkey {
-    my Net::OpenID::Server $self = shift;
-    my $which = (caller(1))[3] =~ /public/ ? "public" : "private";
-
-    if (my $key = shift) {
-        Carp::croak("Too many parameters") if @_;
-        $self->{"${which}_key"} = $key;
-    } else {
-        return $self->_get_key_contents($which);
-    }
-}
-
 sub _fail {
     my Net::OpenID::Server $self = shift;
     $self->{last_errcode} = shift;
@@ -583,66 +565,6 @@ sub errtext {
     $self->{last_errtext};
 }
 
-sub _dsa_sig {
-    my ($self, $plain) = @_;
-    my $msg = sha1($plain);
-
-    if ($HAS_CRYPT_OPENSSL) {
-        my $priv = $self->_get_key_contents("private")
-            or return;
-        my $dsa_priv  = Crypt::OpenSSL::DSA->read_priv_key_str($priv)
-            or $self->_fail("privkey_parse_error", "Couldn't parse private key");
-        return $dsa_priv->sign($msg);
-    }
-
-    if ($HAS_CRYPT_DSA) {
-        my $dsa = Crypt::DSA->new;
-        my $key;
-        if (-f $self->{private_key}) {
-            $key = Crypt::DSA::Key->new(Filename => $self->{private_key},
-                                        Type => 'PEM');
-        } elsif ($Crypt::DSA::VERSION >= 0.14) {
-            my $priv = $self->_get_key_contents("private")
-                or return;
-
-            $key = Crypt::DSA::Key->new(Content => $priv,
-                                        Type => 'PEM');
-        } else {
-            my $priv = $self->_get_key_contents("private")
-                or return;
-
-            # gross:  diving into the innards of Crypt::DSA::Key so
-            # we don't have to write a private key to disk
-            $key = Crypt::DSA::Key->new;
-
-            my $class = 'Crypt::DSA::Key::PEM';
-            eval "use $class;";
-            Carp::croak("Unable to load Crypt::DSA::Key::PEM\n") if $@;
-            bless $key, $class;
-
-            # gross, gross, gross...
-            $key->{Content} = $priv;
-            $key->deserialize(Content => $priv);
-        }
-
-        return $self->_fail("privkey_parse_error", "Couldn't parse private key")
-            unless $key;
-
-        my $sigobj = $dsa->sign(Digest => $msg, Key => $key);
-        return $self->_fail("sign_failure", "Couldn't do Crypt::DSA sign")
-            unless $sigobj && $sigobj->r;
-
-        my $asn = Convert::ASN1->new;
-        $asn->prepare("SEQUENCE { r INTEGER, s INTEGER }");
-        my $sig = $asn->encode( r => $sigobj->r, s => $sigobj->s );
-        return $sig;
-    }
-
-    # OpenSSL mode, if it's later supported:
-    # openssl dgst -dss1 -hex -sign yadis_private.key foobar.txt > sig.txt
-
-}
-
 sub _eurl
 {
     my $a = $_[0];
@@ -704,6 +626,19 @@ sub _time_to_w3c {
                    $hour, $min, $sec);
 }
 
+sub _w3c_to_time {
+    my $hms = shift;
+    return 0 unless
+        $hms =~ /^(\d{4,4})-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)Z$/;
+
+    my $time;
+    eval {
+        $time = timegm($6, $5, $4, $3, $2 - 1, $1);
+    };
+    return 0 if $@;
+    return $time;
+}
+
 sub _push_url_arg {
     my $uref = shift;
     $$uref =~ s/[&?]$//;
@@ -732,9 +667,7 @@ sub _bi2bytes {
 }
 
 sub _bi2arg {
-    my $b64 = MIME::Base64::encode_base64(_bi2bytes($_[0]));
-    chomp $b64;
-    return $b64;
+    return _b64(_bi2bytes($_[0]));
 }
 
 sub _bytes2bi {
@@ -742,15 +675,26 @@ sub _bytes2bi {
 }
 
 sub _arg2bi {
+    return undef unless defined $_[0] && $_[0] ne "";
     # don't acccept base-64 encoded numbers over 700 bytes.  which means
     # those over 4200 bits.
     return Math::BigInt->new("0") if length($_[0]) > 700;
     return _bytes2bi(MIME::Base64::decode_base64($_[0]));
 }
 
+sub _default_p {
+    return Math::BigInt->new("155172898181473697471232257763715539915724801966915404479707795314057629378541917580651227423698188993727816152646631438561595825688188889951272158842675419950341258706556549803580104870537681476726513255747040765857479291291572334510643245094715007229621094194349783925984760375594985848253359305585439638443");
+}
+
+sub _default_g {
+    return Math::BigInt->new("2");
+}
+
 sub _rand_chars
 {
+    shift if @_ == 2;  # shift off classname/obj, if called as method
     my $length = shift;
+
     my $chal = "";
     my $digits = "abcdefghijklmnopqrstuvwzyzABCDEFGHIJKLMNOPQRSTUVWZYZ0123456789";
     for (1..$length) {
@@ -759,6 +703,9 @@ sub _rand_chars
     return $chal;
 }
 
+# also a public interface:
+*rand_chars = \&_rand_chars;
+
 # From Digest::HMAC
 sub hmac_sha1_hex {
     unpack("H*", &hmac_sha1);
@@ -788,7 +735,8 @@ Net::OpenID::Server - library for consumers of OpenID identities
   use Net::OpenID::Server;
 
   my $nos = Net::OpenID::Server->new(
-    args         => $cgi,
+    get_args     => $cgi,
+    post_args    => $cgi,
     private_key  => \&get_priv_key_from_database,
     public_key   => "public_key.txt",
     get_user     => \&get_user,
@@ -828,9 +776,9 @@ identity.  More information is available at:
 =item Net::OpenID::Server->B<new>([ %opts ])
 
 You can set anything in the constructor options that there are
-getters/setters methods for below.  That includes: args, private_key,
-public_key, get_user, is_identity, is_trusted, setup_url, and setup_map.
-See below for docs.
+getters/setters methods for below.  That includes: get_args,
+post_args, private_key, public_key, get_user, is_identity, is_trusted,
+setup_url, and setup_map.  See below for docs.
 
 =back
 
@@ -897,39 +845,51 @@ Required.  The identity URL to sign.
 
 Required.  The base of the URL being generated.
 
-=item C<trust_Root>
+=item C<assoc_handle>
+
+The association handle to use for the signature.  If blank, dumb
+consumer mode is used, and the library picks the handle.
+
+=item C<trust_root>
 
 Optional.  If present, the C<return_to> URL will be checked to be within
 ("under") this trust_root.  If not, the URL returned will be undef.
 
 =back
 
-=item $nos->B<args>($ref)
+=item $nos->B<get_args>($ref)
 
-=item $nos->B<args>($param)
+=item $nos->B<get_args>($param)
 
-=item $nos->B<args>
+=item $nos->B<get_args>
+
+=item $nos->B<post_args>($ref)
+
+=item $nos->B<post_args>($param)
+
+=item $nos->B<post_args>
 
 Can be used in 1 of 3 ways:
 
 1. Setting the way which the Server instances obtains GET parameters:
 
-$nos->args( $reference )
+$nos->get_args( $reference )
 
-Where $reference is either a HASH ref, CODE ref, Apache $r,
-Apache::Request $apreq, or CGI.pm $cgi.  If a CODE ref, the subref
-must return the value given one argument (the parameter to retrieve)
+Where $reference is either a HASH ref, CODE ref, Apache $r (for
+get_args only), Apache::Request $apreq, or CGI.pm $cgi.  If a CODE
+ref, the subref must return the value given one argument (the
+parameter to retrieve)
 
 2. Get a paramater:
 
-my $foo = $nos->args("foo");
+my $foo = $nos->get_args("foo");
 
 When given an unblessed scalar, it retrieves the value.  It croaks if
 you haven't defined a way to get at the parameters.
 
 3. Get the getter:
 
-my $code = $nos->args;
+my $code = $nos->get_args;
 
 Without arguments, returns a subref that returns the value given a
 parameter name.
@@ -1000,7 +960,8 @@ secret based on the provided I<$time>, a unix timestamp.  And if one
 doesn't exist for that time, create, store and return it (with
 appropriate locking so you never return different secrets for the same
 time.)  Your secret can just be random characters, but it's your
-responsibility to do the locking and storage.
+responsibility to do the locking and storage.  If you want help
+generating random characters, call C<Net::OpenID::Server::rand_chars($len)>.
 
 Your secret may not exceed 255 characters.
 
@@ -1021,10 +982,15 @@ When this module gives a consumer site a user_setup_url from your
 provided setup_url, it also has to append a number of get parameters
 onto your setup_url, so your app based at that setup_url knows what it
 has to setup.  Those keys are named, by default, "trust_root",
-"return_to", "post_grant", and "identity".  If you don't like those
-parameter names, this $hashref setup_map lets you change one or more
-of them.  The hashref's keys should be the default values, with values
-being the parameter names you want.
+"return_to", "post_grant", "identity", and "assoc_handle".  If you
+don't like those parameter names, this $hashref setup_map lets you
+change one or more of them.  The hashref's keys should be the default
+values, with values being the parameter names you want.
+
+=item Net::OpenID::Server->rand_chars($len)
+
+Utility function to return a string of $len random characters.  May be
+called as package method, object method, or regular function.
 
 =item $nos->B<err>
 
